import { StringCast } from '../model/casts/StringCast'
import type { ModelFields } from '../model/Model'
import { BooleanCast } from '../model/casts/BooleanCast'
import { NumberCast } from '../model/casts/NumberCast'
import { ArrayCast } from '../model/casts/ArrayCast'
import { CastAttribute } from '../model/casts/CastAttribute'

interface SortableArray<T> {
  criteria: any[]
  index: number
  value: T
}

/**
 * Compare two values with custom string operator
 */
export function compareWithOperator(leftValue: any, rightValue: any, operator?: string): boolean {
  switch (operator) {
    case '>': return leftValue > rightValue
    case '>=': return leftValue >= rightValue
    case '<': return leftValue < rightValue
    case '<=': return leftValue <= rightValue
    case '=': return leftValue === rightValue
    case '!=': return leftValue !== rightValue
    default: return leftValue === rightValue
  }
}

/**
 * Check if the given value is the type of null.
 */
export function isNull(value: any): value is null {
  return value === null
}

/**
 * Check if the given value is the type of undefined or null.
 */
export function isNullish(value: any): value is undefined | null {
  return value === undefined || isNull(value)
}

/**
 * Check if the given value is the type of array.
 */
export function isArray(value: any): value is any[] {
  return Array.isArray(value)
}

/**
 * Check if the given value is the type of array.
 */
export function isFunction(value: any): value is Function {
  return typeof value === 'function'
}

/**
 * Check if the given array or object is empty.
 */
export function isEmpty(collection: any[] | object): boolean {
  return size(collection) === 0
}

/**
 * Gets the size of collection by returning its length for array-like values
 * or the number of own enumerable string keyed properties for objects.
 */
export function size(collection: any[] | object): number {
  return isArray(collection)
    ? collection.length
    : Object.keys(collection).length
}

/**
 * Creates an array of elements, sorted in specified order by the results
 * of running each element in a collection thru each iteratee.
 */
export function orderBy<T>(
  collection: T[],
  iteratees: (((record: T) => any) | string)[],
  directions: string[],
): T[] {
  let index = -1

  const result = collection.map<SortableArray<T>>((value) => {
    const criteria = iteratees.map((iteratee) => {
      return typeof iteratee === 'function' ? iteratee(value) : value[iteratee]
    })

    return { criteria, index: ++index, value }
  })

  return baseSortBy(result, (object, other) => {
    return compareMultiple(object, other, directions)
  })
}

/**
 * Creates an array of elements, sorted in ascending order by the results of
 * running each element in a collection thru each iteratee. This method
 * performs a stable sort, that is, it preserves the original sort order
 * of equal elements.
 */
function baseSortBy<T>(
  array: SortableArray<T>[],
  comparer: (a: SortableArray<T>, B: SortableArray<T>) => number,
): T[] {
  let length = array.length

  array.sort(comparer)

  const newArray: T[] = []

  while (length--) newArray[length] = array[length].value

  return newArray
}

/**
 * Used by `orderBy` to compare multiple properties of a value to another
 * and stable sort them.
 *
 * If `orders` is unspecified, all values are sorted in ascending order.
 * Otherwise, specify an order of "desc" for descending or "asc" for
 * ascending sort order of corresponding values.
 */
function compareMultiple<T>(
  object: SortableArray<T>,
  other: SortableArray<T>,
  directions: string[],
): number {
  let index = -1

  const objCriteria = object.criteria
  const othCriteria = other.criteria
  const length = objCriteria.length

  while (++index < length) {
    const result = compareAscending(objCriteria[index], othCriteria[index])

    if (result) {
      const direction = directions[index]

      return result * (direction === 'desc' ? -1 : 1)
    }
  }

  return object.index - other.index
}

/**
 * Compares values to sort them in ascending order.
 */
function compareAscending(value: any, other: any): number {
  if (value !== other) {
    const valIsDefined = value !== undefined
    const valIsNull = value === null
    const valIsReflexive = value === value

    const othIsDefined = other !== undefined
    const othIsNull = other === null

    if (typeof value !== 'number' || typeof other !== 'number') {
      value = String(value)
      other = String(other)
    }

    if (
      (!othIsNull && value > other)
      || (valIsNull && othIsDefined)
      || !valIsDefined
      || !valIsReflexive
    )
      return 1

    return -1
  }

  return 0
}

/**
 * Creates an object composed of keys generated from the results of running
 * each element of collection through iteratee.
 */
export function groupBy<T>(
  collection: T[],
  iteratee: (record: T) => string,
): { [key: string]: T[] } {
  return collection.reduce((records, record) => {
    const key = iteratee(record)

    if (records[key] === undefined)
      records[key] = []

    records[key].push(record)

    return records
  }, {})
}

export function convertCast(attributes: ModelFields, caster: string | typeof CastAttribute) {
  if (typeof caster !== 'string' && caster)
    return caster.newRawInstance(attributes)

  switch (caster) {
    case 'string': return new StringCast(attributes)
    case 'boolean': return new BooleanCast(attributes)
    case 'number': return new NumberCast(attributes)
    case 'array': return new ArrayCast(attributes)
    default: return new CastAttribute(attributes)
  }
}

/**
 * Asserts that the condition is truthy, throwing immediately if not.
 */
export function throwError(
  message: string[],
): void {
  throw new Error(['[Pinia ORM]'].concat(message).join(' '))
}

/**
 * Asserts that the condition is truthy, throwing immediately if not.
 */
export function assert(
  condition: boolean,
  message: string[],
): asserts condition {
  if (!condition)
    throwError(message)
}
